"""Kernel / covariance function."""

import abc
from typing import Generic, Optional, Tuple, TypeVar, Union

import numpy as np

import probnum.utils as _utils
from probnum.type import IntArgType, ShapeArgType, ShapeType

_InputType = TypeVar("InputType")


class Kernel(Generic[_InputType], abc.ABC):
    """Kernel / covariance function.

    Abstract base class for kernels / covariance functions. Kernels are a
    generalization of a positive-definite function or matrix. They
    typically define the covariance function of a random process and thus describe
    its spatial or temporal variation. If evaluated at two sets of points a kernel
    gives the covariance of the random process at these locations.

    Parameters
    ----------
    input_dim :
        Input dimension of the kernel.
    output_dim :
        Output dimension of the kernel.

    Examples
    --------
    Kernels are implemented by subclassing this abstract base class.

    >>> from probnum.kernels import Kernel
    ...
    >>> class CustomLinearKernel(Kernel):
    ...
    ...     def __init__(self, constant=0.0):
    ...         self.constant = constant
    ...         super().__init__(input_dim=1, output_dim=1)
    ...
    ...     def __call__(self, x0, x1=None):
    ...         # Check and reshape inputs
    ...         x0, x1, kernshape = self._check_and_reshape_inputs(x0, x1)
    ...
    ...         # Compute kernel matrix
    ...         if x1 is None:
    ...             x1 = x0
    ...         kernmat = x0 @ x1.T + self.constant
    ...
    ...         return Kernel._reshape_kernelmatrix(kernmat, newshape=kernshape)

    We can now evaluate the kernel like so.

    >>> import numpy as np
    >>> k = CustomLinearKernel(constant=1.0)
    >>> k(np.linspace(0, 1, 4)[:, None])
    array([[1.        , 1.        , 1.        , 1.        ],
           [1.        , 1.11111111, 1.22222222, 1.33333333],
           [1.        , 1.22222222, 1.44444444, 1.66666667],
           [1.        , 1.33333333, 1.66666667, 2.        ]])
    """

    # pylint: disable="invalid-name"
    def __init__(
        self,
        input_dim: IntArgType,
        output_dim: IntArgType = 1,
    ):
        self._input_dim = np.int_(_utils.as_numpy_scalar(input_dim))
        self._output_dim = np.int_(_utils.as_numpy_scalar(output_dim))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @abc.abstractmethod
    def __call__(
        self, x0: _InputType, x1: Optional[_InputType] = None
    ) -> Union[np.ndarray, np.float_]:
        """Evaluate the kernel.

        Computes the covariance function at ``x0`` and ``x1``. If the inputs have
        more than one dimension the covariance function is evaluated pairwise for all
        observations determined by the first dimension of ``x0`` and ``x1``. If
        only ``x0`` is given the kernel matrix :math:`K=k(X_0, X_0)` is computed.

        Parameters
        ----------
        x0 :
            *shape=(input_dim,) or (n0, input_dim)* -- First input.
        x1 :
            *shape=(input_dim,) or (n1, input_dim)* -- Second input.

        Returns
        -------
        cov :
            *shape=(), (output_dim, output_dim) or (n0, n1) or (n0, n1, output_dim,
            output_dim)* -- Kernel evaluated at ``x0`` and ``x1`` or kernel matrix
            containing pairwise evaluations for all observations in ``x0`` (and ``x1``).
        """
        raise NotImplementedError

    @property
    def input_dim(self) -> int:
        """Dimension of arguments of the covariance function.

        The dimension of inputs to the covariance function :math:`k : \\mathbb{R}^{
        d_{in}} \\times \\mathbb{R}^{d_{in}} \\rightarrow
        \\mathbb{R}^{d_{out} \\times d_{out}}`.
        """
        return self._input_dim

    @property
    def output_dim(self) -> int:
        """Dimension of the evaluated covariance function.

        The resulting evaluated kernel :math:`k(x_0, x_1) \\in
        \\mathbb{R}^{d_{out} \\times d_{out}}` has *shape=(output_dim,
        output_dim)*.
        """
        return self._output_dim

    def _check_and_reshape_inputs(
        self,
        x0: _InputType,
        x1: Optional[_InputType] = None,
    ) -> Tuple[np.ndarray, Optional[np.ndarray], ShapeType]:
        """Check and transform inputs of the covariance function.

        Checks the shape of the inputs to the covariance function and
        transforms the inputs into two-dimensional :class:`numpy.ndarray`s such that
        inputs are stacked row-wise.

        Parameters
        ----------
        x0 :
            First input to the covariance function.
        x1 :
            Second input to the covariance function.

        Returns
        -------
        x0 :
            First input to the covariance function.
        x1 :
            Second input to the covariance function.
        kernshape :
            Shape of the evaluation of the covariance function.

        Raises
        -------
        ValueError :
            If input shapes of x0 and x1 do not match the kernel input dimension or
            each other.
        """
        # pylint: disable="too-many-boolean-expressions"

        # Check and promote shapes
        x0 = np.asarray(x0)
        if x1 is None:
            if (
                (x0.ndim == 0 and self.input_dim > 1)  # Scalar input
                or (x0.ndim == 1 and x0.shape[0] != self.input_dim)  # Vector input
                or (x0.ndim >= 2 and x0.shape[1] != self.input_dim)  # Matrix input
            ):
                raise ValueError(
                    f"Argument shape x0.shape={x0.shape} does not match "
                    "kernel input dimension."
                )

            # Determine correct shape for the kernel matrix as the output of __call__
            kernshape = self._get_shape_kernelmatrix(
                x0_shape=x0.shape, x1_shape=x0.shape
            )

            return np.atleast_2d(x0), None, kernshape
        else:
            x1 = np.asarray(x1)
            err_msg = (
                f"Argument shapes x0.shape={x0.shape} and x1.shape="
                f"{x1.shape} do not match kernel input dimension "
                f"{self.input_dim}. Try passing either two vectors or two "
                "matrices with the second dimension equal to the kernel input "
                "dimension."
            )

            # pylint: disable=redefined-variable-type

            # Promote unequal shapes
            if x0.ndim < 2 and x1.ndim == 2:
                x0 = np.atleast_2d(x0)
            if x1.ndim < 2 and x0.ndim == 2:
                x1 = np.atleast_2d(x1)
            if x0.ndim != x1.ndim:  # Shape mismatch
                raise ValueError(err_msg)

            # Check shapes
            if (
                (x0.ndim == 0 and self.input_dim > 1)  # Scalar input
                or (
                    x0.ndim == 1  # Vector input
                    and not (x0.shape[0] == x1.shape[0] == self.input_dim)
                )
                or (
                    x0.ndim == 2  # Matrix input
                    and not (x0.shape[1] == x1.shape[1] == self.input_dim)
                )
            ):
                raise ValueError(err_msg)

            # Determine correct shape for the kernel matrix as the output of __call__
            kernshape = self._get_shape_kernelmatrix(
                x0_shape=x0.shape, x1_shape=x1.shape
            )

            return np.atleast_2d(x0), np.atleast_2d(x1), kernshape

    def _get_shape_kernelmatrix(
        self,
        x0_shape: ShapeArgType,
        x1_shape: ShapeArgType,
    ) -> ShapeType:
        """Determine the shape of the kernel matrix based on the given arguments.

        Determine the correct shape of the covariance function evaluated at the given
        input arguments. If inputs are vectors the output is a numpy scalar if the
        output dimension of the kernel is 1, otherwise *shape=(output_dim,
        output_dim)*. If inputs represent multiple observations, then the resulting
        matrix has *shape=(n0, n1) or (n0, n1, output_dim, output_dim)*.

        Parameters
        ----------
        x0_shape :
            Shape of the first input to the covariance function.
        x1_shape :
            Shape of the second input to the covariance function.
        """
        if len(x0_shape) <= 1 and len(x1_shape) <= 1:
            if self.output_dim == 1:
                kern_shape = 0
            else:
                kern_shape = ()
        else:
            kern_shape = (x0_shape[0], x1_shape[0])

        if self.output_dim > 1:
            kern_shape += (
                self.output_dim,
                self.output_dim,
            )

        return _utils.as_shape(kern_shape)

    @staticmethod
    def _reshape_kernelmatrix(
        kerneval: np.ndarray, newshape: ShapeArgType
    ) -> np.ndarray:
        """Reshape the evaluation of the covariance function.

        Reshape the given evaluation of the covariance function to the correct shape,
        determined by the inputs x0 and x1. This method is designed to be called by
        subclasses of :class:`Kernel` in their :meth:`__call__` function to ensure
        the returned quantity has the correct shape independent of the implementation of
        the kernel.

        Parameters:
        -----------
        kerneval
            Covariance function evaluated at ``x0`` and ``x1``.
        newshape :
            New shape of the evaluation of the covariance function.
        """
        if newshape[0] == 0:
            return _utils.as_numpy_scalar(kerneval.squeeze())
        else:
            return kerneval.reshape(newshape)



"""Policies for 1D quadratic optimization."""

from typing import Callable, Optional

import numpy as np

from probnum import randvars
from probnum.type import FloatArgType, RandomStateArgType


def explore_exploit_policy(
    fun: Callable[[FloatArgType], FloatArgType],
    fun_params0: randvars.RandomVariable,
    random_state: RandomStateArgType = None,
) -> float:
    """Policy exploring around the estimate of the minimum based on the certainty about
    the parameters.

    Parameters
    ----------
    fun :
        One-dimensional objective function.
    fun_params0 :
        Belief over the parameters of the quadratic objective.
    random_state :
        Random state of the policy. If None (or np.random), the global np.random state
        is used. If integer, it is used to seed the local
        :class:`~numpy.random.RandomState` instance.
    """
    a0, b0, _ = fun_params0
    return (
        -b0.mean / a0.mean
        + randvars.Normal(
            0, np.trace(fun_params0.cov), random_state=random_state
        ).sample()
    )


def stochastic_policy(
    fun: Callable[[FloatArgType], FloatArgType],
    fun_params0: randvars.RandomVariable,
    random_state: RandomStateArgType = None,
) -> float:
    """Policy returning a random action.

    Parameters
    ----------
    fun :
        One-dimensional objective function.
    fun_params0 :
        Belief over the parameters of the quadratic objective.
    random_state :
        Random state of the policy. If None (or np.random), the global np.random state
        is used. If integer, it is used to seed the local
        :class:`~numpy.random.RandomState` instance.
    """
    return randvars.Normal(mean=0.0, cov=1.0, random_state=random_state).sample()



"""Utility functions for argument types."""

import numbers
from typing import Optional

import numpy as np
import scipy._lib._util

from probnum.type import (
    DTypeArgType,
    RandomStateArgType,
    RandomStateType,
    ScalarArgType,
    ShapeArgType,
    ShapeType,
)

__all__ = ["as_shape", "as_random_state", "as_numpy_scalar"]


def as_random_state(seed: RandomStateArgType) -> RandomStateType:
    """Turn ``seed`` into a np.random.RandomState instance.

    Parameters
    ----------
    seed
        If seed is None, return the RandomState singleton used by np.random. If seed is
        an int, return a new RandomState instance seeded with seed. If seed is already a
        RandomState instance, return it.

    Raises
    -------
    ValueError
        If seed is neither None, an int or a RandomState instance.
    """
    return scipy._lib._util.check_random_state(seed)


def as_shape(x: ShapeArgType, ndim: Optional[numbers.Integral] = None) -> ShapeType:
    """Convert a shape representation into a shape defined as a tuple of ints.

    Parameters
    ----------
    x
        Shape representation.
    """
    if isinstance(x, (int, numbers.Integral, np.integer)):
        shape = (int(x),)
    elif isinstance(x, tuple) and all(isinstance(item, int) for item in x):
        shape = x
    else:
        try:
            _ = iter(x)
        except TypeError as e:
            raise TypeError(
                f"The given shape {x} must be an integer or an iterable of integers."
            ) from e

        if not all(isinstance(item, (int, numbers.Integral, np.integer)) for item in x):
            raise TypeError(f"The given shape {x} must only contain integer values.")

        shape = tuple(int(item) for item in x)

    if isinstance(ndim, numbers.Integral):
        if len(shape) != ndim:
            raise TypeError(f"The given shape {shape} must have {ndim} dimensions.")

    return shape


def as_numpy_scalar(x: ScalarArgType, dtype: DTypeArgType = None) -> np.generic:
    """Convert a scalar into a NumPy scalar.

    Parameters
    ----------
    x
        Scalar value.
    dtype
        Data type of the scalar.
    """

    if np.ndim(x) != 0:
        raise ValueError("The given input is not a scalar.")

    return np.asarray(x, dtype=dtype)[()]



import numpy as np
import pytest_cases

from probnum import _randomvariablelist, diffeq, randvars


@pytest_cases.fixture
@pytest_cases.parametrize_with_cases(
    "testsolver, scipysolver", cases=".test_wrappedscipy_cases"
)
def solution_case(testsolver, scipysolver):
    testsolution = testsolver.solve(diffeq.AdaptiveSteps(0.1, atol=1e-12, rtol=1e-12))
    scipysolution = testsolution.scipy_solution
    return testsolution, scipysolution


def test_locations(solution_case):
    testsolution, scipysolution = solution_case
    scipy_t = scipysolution.ts
    probnum_t = testsolution.locations
    np.testing.assert_allclose(scipy_t, probnum_t, atol=1e-14, rtol=1e-14)


def test_call_isscalar(solution_case):
    testsolution, scipysolution = solution_case
    t = 0.1
    call_scalar = testsolution(t)
    call_array = testsolution([0.1, 0.2, 0.3])
    assert np.isscalar(t)
    assert isinstance(call_scalar, randvars.Constant)
    assert isinstance(call_array, _randomvariablelist._RandomVariableList)


def test_states(solution_case):
    testsolution, scipysolution = solution_case
    scipy_states = np.array(scipysolution(scipysolution.ts)).T
    probnum_states = np.array(testsolution.states.mean)
    np.testing.assert_allclose(scipy_states, probnum_states, atol=1e-14, rtol=1e-14)


def test_call__(solution_case):
    testsolution, scipysolution = solution_case
    scipy_call = scipysolution(scipysolution.ts)
    probnum_call = testsolution(scipysolution.ts).mean.T
    np.testing.assert_allclose(scipy_call, probnum_call, atol=1e-14, rtol=1e-14)



import numpy as np
import pytest

import probnum.problems.zoo.filtsmooth as filtsmooth_zoo
from probnum import filtsmooth, randvars


def test_effective_number_of_events():
    weights = np.random.rand(10)
    categ = randvars.Categorical(
        support=np.random.rand(10, 2), probabilities=weights / np.sum(weights)
    )
    ess = filtsmooth.effective_number_of_events(categ)
    assert 0 < ess < 10


#####################################
# Test the RMSE on a pendulum example
#####################################

# Measmod style checks bootstrap and Gaussian proposals.
all_importance_distributions = pytest.mark.parametrize("measmod_style", ["ukf", "none"])

# Resampling percentage threshold checks that
# resampling is performed a) never, b) sometimes, c) always
all_resampling_configurations = pytest.mark.parametrize(
    "resampling_percentage_threshold", [-1.0, 0.1, 2.0]
)


@pytest.fixture
def num_particles():
    return 10


@pytest.fixture
def problem():
    return filtsmooth_zoo.pendulum(step=0.12)


@pytest.fixture
def particle_filter(
    problem, num_particles, measmod_style, resampling_percentage_threshold
):
    _, statespace_components = problem
    linearized_measmod = (
        filtsmooth.DiscreteUKFComponent(statespace_components["measurement_model"])
        if measmod_style == "ukf"
        else None
    )

    particle = filtsmooth.ParticleFilter(
        statespace_components["dynamics_model"],
        statespace_components["measurement_model"],
        statespace_components["initrv"],
        num_particles=num_particles,
        linearized_measurement_model=linearized_measmod,
        resampling_percentage_threshold=resampling_percentage_threshold,
    )
    return particle


@pytest.fixture()
def regression_problem(problem):
    """Filter and regression problem."""
    regression_problem, _ = problem

    return regression_problem


@all_importance_distributions
@all_resampling_configurations
def test_random_state(particle_filter):
    initrv = particle_filter.initrv
    assert initrv.random_state == particle_filter.random_state


@pytest.fixture
def pf_output(particle_filter, regression_problem):
    posterior, _ = particle_filter.filter(regression_problem)
    return posterior


@all_importance_distributions
@all_resampling_configurations
def test_shape_pf_output(pf_output, regression_problem, num_particles):
    np.random.seed(12345)

    states = pf_output.states.support
    weights = pf_output.states.probabilities
    num_gridpoints = len(regression_problem.locations)
    assert states.shape == (num_gridpoints, num_particles, 2)
    assert weights.shape == (num_gridpoints, num_particles)


@all_importance_distributions
@all_resampling_configurations
def test_rmse_particlefilter(pf_output, regression_problem):
    """Assert that the RMSE of the mode of the posterior of the PF is a lot smaller than
    the RMSE of the data."""

    np.random.seed(12345)

    true_states = regression_problem.solution

    mode = pf_output.states.mode
    rmse_mode = np.linalg.norm(np.sin(mode) - np.sin(true_states)) / np.sqrt(
        true_states.size
    )
    rmse_data = np.linalg.norm(
        regression_problem.observations - np.sin(true_states)
    ) / np.sqrt(true_states.size)

    # RMSE of PF.mode strictly better than RMSE of data
    assert rmse_mode < 0.9 * rmse_data



"""White noise kernel."""

from typing import Optional

import numpy as np

import probnum.utils as _utils
from probnum.type import IntArgType, ScalarArgType

from ._kernel import Kernel

_InputType = np.ndarray


class WhiteNoise(Kernel[_InputType]):
    """White noise kernel.

    Kernel representing independent and identically distributed white noise :math:`k(
    x_0, x_1) = \\sigma^2 \\delta(x_0, x_1)`.

    Parameters
    ----------
    input_dim :
        Input dimension of the kernel.
    sigma :
        Noise level.
    """

    def __init__(self, input_dim: IntArgType, sigma: ScalarArgType = 1.0):
        self.sigma = _utils.as_numpy_scalar(sigma)
        super().__init__(input_dim=input_dim, output_dim=1)

    def __call__(self, x0: _InputType, x1: Optional[_InputType] = None) -> np.ndarray:

        x0, x1, kernshape = self._check_and_reshape_inputs(x0, x1)

        # Compute kernel matrix
        if x1 is None:
            kernmat = self.sigma ** 2 * np.eye(x0.shape[0])
        else:
            kernmat = self.sigma ** 2 * np.equal(x0, x1[:, np.newaxis, :]).all(
                axis=2
            ).T.astype(float)

        return Kernel._reshape_kernelmatrix(kernmat, newshape=kernshape)



"""Particle filters."""

from typing import Optional, Union

import numpy as np

from probnum import problems, randvars, statespace
from probnum.filtsmooth.bayesfiltsmooth import BayesFiltSmooth
from probnum.type import FloatArgType, IntArgType

from ._particle_filter_posterior import ParticleFilterPosterior


def effective_number_of_events(categ_rv: randvars.Categorical) -> float:
    """Approximate effective number of events in the support of a categorical random
    variable.

    In a particle filter, this is used as the effective number of
    particles which may indicate the need for resampling.
    """
    return 1.0 / np.sum(categ_rv.probabilities ** 2)


class ParticleFilter(BayesFiltSmooth):
    r"""Particle filter (PF). Also known as sequential Monte Carlo method.

    A PF estimates the posterior distribution of a Markov process given noisy, non-linear observations,
    with a set of particles.

    The random state of the particle filter is inferred from the random state of the initial random variable.

    Parameters
    ----------
    dynamics_model :
        Prior dynamics. Since the PF is essentially a discrete-time algorithm,
        the prior must be a discrete model (or at least one with an equivalent discretisation).
        This transition must support `forward_realization`.
    measurement_model :
        Measurement model. Must be a discrete model that supports `forward_realization`.
    initrv :
        Initial random variable. Can be any `RandomVariable` object that implements `sample()`.
    num_particles :
        Number of particles to use.
    linearized_measurement_model :
        Linearized measurement model that is used as an importance density. In principle,
        any discrete-time model that supports `backward_realization` is applicable.
        In practice, it will almost always be one out of `DiscreteEKFComponent`, `DiscreteUKFComponent`,
        or `IteratedDiscreteComponent`. Linear components are also possible, but would most often imply
        that a particle filter is not required, because the filtering problem can be used much faster
        with a Kalman filter. The exception to this rule is if the initial random variable is not Gaussian.
        Optional. Default is None, which implies the bootstrap PF.
    with_resampling :
        Whether after each step the effective number of particles shall be checked, and, if too low,
        the state should be resampled. Optional. Default is `True`.
    resampling_percentage_threshold :
        Percentage threshold for resampling. That is, it is the value :math:`p` such that
        resampling is performed if :math:`N_{\text{eff}} < p \, N_\text{particles}` holds.
        Optional. Default is 0.1. If this value is non-positive, resampling is never performed.
        If it is larger than 1, resampling is performed after each step.
    """

    def __init__(
        self,
        dynamics_model: Union[statespace.LTISDE, statespace.DiscreteGaussian],
        measurement_model: statespace.DiscreteGaussian,
        initrv: randvars.RandomVariable,
        num_particles: IntArgType,
        linearized_measurement_model: Optional[statespace.DiscreteGaussian] = None,
        with_resampling: bool = True,
        resampling_percentage_threshold: FloatArgType = 0.1,
    ) -> None:
        super().__init__(
            dynamics_model=dynamics_model,
            measurement_model=measurement_model,
            initrv=initrv,
        )
        self.num_particles = num_particles

        self.with_resampling = with_resampling
        self.resampling_percentage_threshold = resampling_percentage_threshold
        self.min_effective_num_of_particles = (
            resampling_percentage_threshold * num_particles
        )

        # If None, the dynamics model is used as a fallback option
        # which results in the bootstrap PF.
        # Any linearised measurement model that could be used in a
        # Gaussian filter can be used here and will likely be a better
        # choice than the bootstrap.
        self.linearized_measurement_model = linearized_measurement_model

    def filter(self, regression_problem: problems.RegressionProblem):
        """Apply particle filtering to a data set.

        Parameters
        ----------
        regression_problem

        Returns
        -------
        posterior
            Posterior distribution of the filtered output
        info_dicts
            list of dictionaries containing filtering information

        See Also
        --------
        RegressionProblem: a regression problem data class
        """
        filtered_rvs = []
        info_dicts = []

        for rv, info in self.filter_generator(regression_problem):
            filtered_rvs.append(rv)
            info_dicts.append(info)

        posterior = ParticleFilterPosterior(
            states=filtered_rvs,
            locations=regression_problem.locations,
        )

        return posterior, info_dicts

    def filter_generator(self, regression_problem: problems.RegressionProblem):
        """Apply Particle filtering to a data set.

        Parameters
        ----------
        regression_problem

        Yields
        ------
        curr_rv
            Filtering random variable at each grid point.
        info_dict
            Dictionary containing filtering information

        See Also
        --------
        RegressionProblem: a regression problem data class
        """

        dataset, times = regression_problem.observations, regression_problem.locations

        # Initialize: condition on first data point using the initial random
        # variable as a dynamics rv (i.e. as the current prior).
        particles_and_weights = np.array(
            [
                self.compute_new_particle(dataset[0], times[0], self.initrv)
                for _ in range(self.num_particles)
            ],
            dtype=object,
        )
        particles = np.stack(particles_and_weights[:, 0], axis=0)
        weights = np.stack(particles_and_weights[:, 1], axis=0)
        weights = np.array(weights) / np.sum(weights)
        curr_rv = randvars.Categorical(
            support=particles, probabilities=weights, random_state=self.random_state
        )
        yield curr_rv, {}

        for idx in range(1, len(times)):
            curr_rv, info_dict = self.filter_step(
                start=times[idx - 1],
                stop=times[idx],
                randvar=curr_rv,
                data=dataset[idx],
            )
            yield curr_rv, info_dict

    def filter_step(self, start, stop, randvar, data):
        """Perform a particle filter step.

        This method implements sequential importance (re)sampling.

        It consists of the following steps:
        1. Propagating the "past" particles through the dynamics model.
        2. Computing a "proposal" random variable.
        This is either the prior dynamics model or the output of a filter step
        of an (approximate) Gaussian filter.
        3. Sample from the proposal random variable. This is the "new" particle.
        4. Propagate the particle through the measurement model.
        This is required in order to evaluate the PDF of the resulting RV at
        the data. If this is small, the weight of the particle will be small.
        5. Compute weights ("event probabilities") of the new particle.
        This requires evaluating the PDFs of all three RVs (dynamics, proposal, measurement).

        After this is done for all particles, the weights are normalized in order to sum to 1.
        If the effective number of particles is low, the particles are resampled.
        """
        new_weights = randvar.probabilities.copy()
        new_support = randvar.support.copy()

        for idx, (particle, weight) in enumerate(zip(new_support, new_weights)):

            dynamics_rv, _ = self.dynamics_model.forward_realization(
                particle, t=start, dt=(stop - start)
            )
            proposal_state, proposal_weight = self.compute_new_particle(
                data, stop, dynamics_rv
            )
            new_support[idx] = proposal_state
            new_weights[idx] = proposal_weight

        new_weights = new_weights / np.sum(new_weights)
        new_rv = randvars.Categorical(
            support=new_support,
            probabilities=new_weights,
            random_state=self.random_state,
        )

        if self.with_resampling:
            if effective_number_of_events(new_rv) < self.min_effective_num_of_particles:
                new_rv = new_rv.resample()

        return new_rv, {}

    def compute_new_particle(self, data, stop, dynamics_rv):
        """Compute a new particle.

        Turn the dynamics RV into a proposal RV, apply the measurement
        model and compute new weights via the respective PDFs.
        """
        proposal_rv = self.dynamics_to_proposal_rv(dynamics_rv, data=data, t=stop)
        proposal_state = proposal_rv.sample()
        meas_rv, _ = self.measurement_model.forward_realization(proposal_state, t=stop)

        # For the bootstrap PF, the dynamics and proposal PDFs cancel out.
        # Therefore we make the following exception.
        if self.linearized_measurement_model is None:
            log_proposal_weight = meas_rv.logpdf(data)
        else:
            log_proposal_weight = (
                meas_rv.logpdf(data)
                + dynamics_rv.logpdf(proposal_state)
                - proposal_rv.logpdf(proposal_state)
            )

        proposal_weight = np.exp(log_proposal_weight)
        return proposal_state, proposal_weight

    def dynamics_to_proposal_rv(self, dynamics_rv, data, t):
        """Turn a dynamics RV into a proposal RV.

        The output of this function depends on the choice of PF. For the
        bootstrap PF, nothing happens. For other PFs, the importance
        density is used to improve the proposal. Currently, only
        approximate Gaussian importance densities are provided.
        """
        proposal_rv = dynamics_rv
        if self.linearized_measurement_model is not None:
            proposal_rv, _ = self.linearized_measurement_model.backward_realization(
                data, proposal_rv, t=t
            )
        return proposal_rv

    @property
    def random_state(self):
        """Random state of the particle filter.

        Inferred from the random state of the initial random variable.
        """
        return self.initrv.random_state



import collections.abc
from functools import partial
from typing import Callable, Dict, Iterable, Optional, Tuple, Union

import numpy as np

import probnum as pn
import probnum.utils as _utils
from probnum import linops, randvars
from probnum.type import FloatArgType, IntArgType, RandomStateArgType

from .belief_updates import gaussian_belief_update
from .observation_operators import function_evaluation
from .policies import explore_exploit_policy, stochastic_policy
from .stopping_criteria import maximum_iterations, parameter_uncertainty

# Type aliases for quadratic optimization
QuadOptPolicyType = Callable[
    [
        Callable[[FloatArgType], FloatArgType],
        randvars.RandomVariable,
    ],
    FloatArgType,
]
QuadOptObservationOperatorType = Callable[
    [Callable[[FloatArgType], FloatArgType], FloatArgType], FloatArgType
]
QuadOptBeliefUpdateType = Callable[
    [
        randvars.RandomVariable,
        FloatArgType,
        FloatArgType,
    ],
    randvars.RandomVariable,
]
QuadOptStoppingCriterionType = Callable[
    [Callable[[FloatArgType], FloatArgType], randvars.RandomVariable, IntArgType],
    Tuple[bool, Union[str, None]],
]


class ProbabilisticQuadraticOptimizer:
    """Probabilistic Quadratic Optimization in 1D.

    PN method solving unconstrained one-dimensional (noisy) quadratic
    optimization problems only needing access to function evaluations.

    Parameters
    ----------
    fun_params_prior :
        Prior belief over the parameters of the latent quadratic function.
    policy :
        Callable returning a new action to probe the problem.
    observation_operator :
        Callable implementing the observation process of the problem.
    belief_update :
        Belief update function updating the belief over the parameters of the quadratic given
        an action and observation of the problem.
    stopping_criteria :
        Stopping criteria to determine when to stop the optimization.

    See Also
    --------
    probsolve_qp : Solve 1D (noisy) quadratic optimization problems.

    Examples
    --------
    >>> from functools import partial
    >>> import numpy as np
    >>> from probnum import randvars
    >>> from quadopt_example.policies import stochastic_policy
    >>> from quadopt_example.observation_operators import function_evaluation
    >>> from quadopt_example.belief_updates import gaussian_belief_update
    >>> from quadopt_example.stopping_criteria import maximum_iterations
    >>>
    >>> np.random.seed(42)
    >>>
    >>> # Custom stopping criterion based on residual
    >>> def residual(
    >>>     fun,
    >>>     fun_params0,
    >>>     current_iter,
    >>>     abstol=10 ** -6,
    >>>     reltol=10 ** -6,
    >>> ):
    >>>     a, b, c = fun_params0.mean
    >>>     resid = np.abs(fun(1.0) - (0.5 * a + b + c))
    >>>     if resid < abstol:
    >>>         return True, "residual_abstol"
    >>>     elif resid < np.abs(fun(1.0)) * reltol:
    >>>         return True, "residual_reltol"
    >>>     else:
    >>>         return False, None
    >>>
    >>> # Compose custom PN method
    >>> quadopt = ProbabilisticQuadraticOptimizer(
    >>>     fun_params_prior=randvars.Normal(np.zeros(3), np.eye(3)),
    >>>     policy=stochastic_policy,
    >>>     observation_operator=function_evaluation,
    >>>     belief_update=partial(gaussian_belief_update, noise_cov=np.zeros(3)),
    >>>     stopping_criteria=[residual, partial(maximum_iterations, maxiter=10)],
    >>> )
    >>> # Objective function
    >>> f = lambda x: 2.0 * x ** 2 - 0.75 * x + 0.2
    >>>
    >>> quadopt.optimize(f)
    (0.2000000000000014,
     <() Normal with dtype=float64>,
     <(3,) Normal with dtype=float64>,
     {'iter': 3, 'conv_crit': 'residual_abstol'})
    """

    def __init__(
        self,
        fun_params_prior: randvars.RandomVariable,
        policy: QuadOptPolicyType,
        observation_operator: QuadOptObservationOperatorType,
        belief_update: QuadOptBeliefUpdateType,
        stopping_criteria: Union[
            QuadOptStoppingCriterionType, Iterable[QuadOptStoppingCriterionType]
        ],
    ):
        # Optimizer components
        self.fun_params = fun_params_prior
        self.policy = policy
        self.observation_operator = observation_operator
        self.belief_update = belief_update

        if not isinstance(stopping_criteria, collections.abc.Iterable):
            self.stopping_criteria = [stopping_criteria]
        else:
            self.stopping_criteria = stopping_criteria

    def has_converged(
        self, fun: Callable[[FloatArgType], FloatArgType], iteration: IntArgType
    ) -> Tuple[bool, Union[str, None]]:
        """Check whether the optimizer has converged.

        Parameters
        ----------
        fun :
            Quadratic objective function to optimize.
        iteration :
            Number of iterations of the solver performed up to this point.
        """
        for stopping_criterion in self.stopping_criteria:
            _has_converged, convergence_criterion = stopping_criterion(
                fun, self.fun_params, iteration
            )
            if _has_converged:
                return True, convergence_criterion
        return False, None

    def optim_iterator(
        self,
        fun: Callable[[FloatArgType], FloatArgType],
    ) -> Tuple[float, float, randvars.RandomVariable]:
        """Generator implementing the optimization iteration.

        This function allows stepping through the optimization
        process one step at a time.

        Parameters
        ----------
        fun :
            Quadratic objective function to optimize.

        Returns
        -------
        action :
            Action to probe the problem.
        observation :
            Observation of the problem for the given ``action``.
        fun_params :
            Belief over the parameters of the objective function.
        """
        while True:
            # Compute action via policy
            action = self.policy(fun, self.fun_params)

            # Make an observation
            observation = self.observation_operator(fun, action)

            # Belief update
            self.fun_params = self.belief_update(self.fun_params, action, observation)

            yield action, observation, self.fun_params

    def optimize(
        self,
        fun: Callable[[FloatArgType], FloatArgType],
        callback: Optional[
            Callable[[float, float, randvars.RandomVariable], None]
        ] = None,
    ) -> Tuple[float, randvars.RandomVariable, randvars.RandomVariable, Dict]:
        """Optimize the quadratic objective function.

        Parameters
        ----------
        fun :
            Quadratic objective function to optimize.
        callback :
            Callback function returning intermediate quantities of the
            optimization loop. Note that depending on the function
            supplied, this can slow down the solver considerably.

        Returns
        -------
        x_opt :
            Estimated minimum of ``fun``.
        fun_opt :
            Belief over the optimal value of the objective function.
        fun_params :
            Belief over the parameters of the objective function.
        info :
            Additional information about the optimization, e.g. convergence.
        """
        # Setup
        _has_converged = False
        iteration = 0
        optimization_iterator = self.optim_iterator(fun=fun)

        # Evaluate stopping criteria
        _has_converged, conv_crit = self.has_converged(fun=fun, iteration=iteration)

        while not _has_converged:

            # Perform one iteration of the optimizer
            action, observation, _ = next(optimization_iterator)

            # Callback function
            if callback is not None:
                callback(action, observation, self.fun_params)

            iteration += 1

            # Evaluate stopping criteria
            _has_converged, conv_crit = self.has_converged(fun=fun, iteration=iteration)

        # Belief over optimal function value and optimum
        x_opt, fun_opt = self.belief_optimum()

        # Information (e.g. on convergence)
        info = {"iter": iteration, "conv_crit": conv_crit}

        return x_opt, fun_opt, self.fun_params, info

    def belief_optimum(self) -> Tuple[float, randvars.RandomVariable]:
        """Compute the belief over the optimum and optimal function value.

        Returns
        -------
        x_opt :
            Estimated minimum of ``fun``.
        fun_opt :
            Belief over the optimal value of the objective function.
        """
        x_opt = -self.fun_params.mean[1] / self.fun_params.mean[0]
        fun_opt = np.array([0.5 * x_opt ** 2, x_opt, 1]).T @ self.fun_params
        return x_opt, fun_opt



from pathlib import Path

TOXINI_FILE = Path("tox.ini")
PYPROJECTTOML_FILE = Path("pyproject.toml")
GLOBAL_DISABLES = {
    "invalid-name",
    "fixme",
    "bad-continuation",
    "no-else-raise",
    "no-else-return",
    "no-member",
    "redefined-variable-type",
}


# Parse ./tox.ini
tox_lines = TOXINI_FILE.read_text().splitlines()
tox_pylint_lines = [
    l for l in tox_lines if l.strip().startswith("pylint") and "--disable" in l
]
tox_disables = {i for l in tox_pylint_lines for i in l.split('"')[1].split(",")}


# Parse ./pyproject.toml
pyproject_lines = PYPROJECTTOML_FILE.read_text().splitlines()
is_in_disable_string = False
pyproject_disables = set()
for line in pyproject_lines:
    if is_in_disable_string and line.startswith('"""'):
        is_in_disable_string = False

    if is_in_disable_string and line:
        pyproject_disables.add(line.strip().strip(","))

    if line.startswith("disable"):
        is_in_disable_string = True


# Check correctness and raise errors
try:
    in_pyproject_butnot_tox = pyproject_disables.difference(tox_disables).difference(
        GLOBAL_DISABLES
    )
    assert not in_pyproject_butnot_tox
except AssertionError:
    raise Exception(
        f"""
    The following pylint messages seem to be disabled in `./pyproject.toml`, but not in
    any of the pylint calls in `./tox.ini`: {in_pyproject_butnot_tox}. This could have
    two reasons with different solutions:
    1. You fixed one or many pylint errors in one of the modules and there is no module
       left that needs this specific message disabled. Then, please also remove it from
       the `./pyproject.toml` file such that pylint uses the most up-to-date
       configuration file.
    2. You added a new global exception to `./pyproject.toml` after deciding that this
       is a message that we do not want to enforce anywhere currently. Then, please add
       this exception also to the `GLOBAL_DISABLES` variable in this this python script
       (`./github/workflows/pylint_check.py`).

    If you are not sure what exactly you are supposed to do, or if you think that this
    message is wrong please feel free to ping @nathanaelbosch.
    """
    )


try:
    in_tox_butnot_pyproject = tox_disables.difference(pyproject_disables)
    assert not in_tox_butnot_pyproject
except AssertionError:
    raise Exception(
        f"""
    The following pylint messages seem to be disabled in `./tox.ini`, but not in
    `./pyproject.toml`: {in_tox_butnot_pyproject}. Please make sure to add them to
    `./pyproject.toml` such that pylint does not raise any warnings that are not
    supposed to be fixed right now.

    If you are not sure what exactly you are supposed to do, or if you think that this
    message is wrong please feel free to ping @nathanaelbosch.
    """
    )


print(
    "The pylint exceptions in `./tox.ini` and `./pyproject.toml` seem to be correctly synchronized."
)



"""Test fixtures for kernels."""

from typing import Dict

import numpy as np
import pytest

import probnum.kernels as kernels
import probnum.quad._integration_measures as measures
from probnum.quad.kernel_embeddings._kernel_embedding import KernelEmbedding

# pylint: disable=unnecessary-lambda


@pytest.fixture(
    params=[pytest.param(seed, id=f"seed{seed}") for seed in range(1)],
    name="random_state",
)
def fixture_random_state(request):
    """Random state(s) used for test parameterization."""
    return np.random.RandomState(seed=request.param)


@pytest.fixture(
    params=[pytest.param(num_data, id=f"ndata{num_data}") for num_data in [1, 2, 20]],
    name="num_data",
)
def fixture_num_data(request) -> int:
    """Size of the dataset."""
    return request.param


@pytest.fixture(
    params=[
        pytest.param(input_dim, id=f"dim{input_dim}") for input_dim in [1, 2, 3, 5]
    ],
    name="input_dim",
)
def fixture_input_dim(request) -> int:
    """Input dimension of the covariance function."""
    return request.param


# Datasets
@pytest.fixture(name="x")
def fixture_x(
    input_dim: int, num_data: int, random_state: np.random.RandomState
) -> np.ndarray:
    """Random data from a standard normal distribution."""
    return random_state.normal(0, 1, size=(num_data, input_dim))


# Measures
@pytest.fixture(
    params=[
        pytest.param(diagonal, id="covdiag" if diagonal else "covfull")
        for diagonal in [True, False]
    ],
    name="cov_diagonal",
)
def fixture_diagonal(request) -> str:
    """Possible forms of the covariance matrix."""
    return request.param


@pytest.fixture(
    params=[pytest.param(name, id=name) for name in ["gauss", "lebesgue"]],
    name="measure_name",
)
def fixture_measure_names(request) -> str:
    """Pedestrian way to deal with integration measures."""
    return request.param


@pytest.fixture(name="measure_params")
def fixture_measure_params(
    measure_name: str,
    input_dim: int,
    cov_diagonal: bool,
    random_state: np.random.RandomState,
) -> Dict:
    params = {"name": measure_name}

    if measure_name == "gauss":
        # set up mean and covariance
        if input_dim == 1:
            mean = random_state.normal(0, 1)
            cov = random_state.uniform(0.5, 1.5)
        else:
            mean = random_state.normal(0, 1, size=(input_dim, 1))
            if cov_diagonal:
                cov = random_state.uniform(0.5, 1.5, size=(input_dim, 1))
            else:
                mat = random_state.normal(0, 1, size=(input_dim, input_dim))
                cov = mat @ mat.T

        params["mean"] = mean
        params["cov"] = cov

    elif measure_name == "lebesgue":
        # set up bounds
        rv = random_state.uniform(0, 1, size=(input_dim, 2))
        domain = (rv[:, 0] - 1.0, rv[:, 1] + 1.0)

        params["domain"] = domain
        params["normalized"] = True

    return params


@pytest.fixture(name="measure")
def fixture_measure(measure_params) -> measures.IntegrationMeasure:
    """Kernel / covariance function."""
    name = measure_params.pop("name")

    if name == "gauss":
        return measures.GaussianMeasure(**measure_params)
    elif name == "lebesgue":
        return measures.LebesgueMeasure(**measure_params)
    raise NotImplementedError


# Kernels
@pytest.fixture(
    params=[
        pytest.param(kerndef, id=kerndef[0].__name__)
        for kerndef in [
            (kernels.ExpQuad, {"lengthscale": 1.25}),
        ]
    ],
    name="kernel",
)
def fixture_kernel(request, input_dim: int) -> kernels.Kernel:
    """Kernel / covariance function."""
    return request.param[0](**request.param[1], input_dim=input_dim)


# Kernel Embeddings
@pytest.fixture(name="kernel_embedding")
def fixture_kernel_embedding(
    request, kernel: kernels.Kernel, measure: measures.IntegrationMeasure
) -> KernelEmbedding:
    """Set up kernel embedding."""
    return KernelEmbedding(kernel, measure)


# Test functions
@pytest.fixture(
    params=[
        pytest.param(fun, id=f"f={key}")
        for key, fun in {
            "x": lambda x: x,
            "x**2": lambda x: x ** 2,
            "sin(x)": lambda x: np.sin(x),
        }.items()
    ],
    name="f1d",
)
def fixture_f1d(request):
    """1D test function for BQ."""
    return request.param



import unittest

import numpy as np

from probnum import randvars
from probnum.filtsmooth.gaussfiltsmooth import unscentedtransform


class TestUnscentedTransform(unittest.TestCase):
    def setUp(self):
        self.ndim = np.random.randint(1, 33)  # 1 < random int < 33
        alpha, beta, kappa = np.random.rand(3)
        self.ut = unscentedtransform.UnscentedTransform(self.ndim, alpha, beta, kappa)
        self.mean = np.random.rand(self.ndim)
        cvr = np.random.rand(self.ndim, self.ndim)
        self.covar = cvr @ cvr.T

    def test_weights_shape(self):
        self.assertEqual(self.ut.mweights.ndim, 1)
        self.assertEqual(self.ut.mweights.shape[0], 2 * self.ndim + 1)
        self.assertEqual(self.ut.cweights.ndim, 1)
        self.assertEqual(self.ut.cweights.shape[0], 2 * self.ndim + 1)

    def test_sigpts_shape(self):
        sigpts = self.ut.sigma_points(randvars.Normal(self.mean, self.covar))
        self.assertEqual(sigpts.ndim, 2)
        self.assertEqual(sigpts.shape[0], 2 * self.ndim + 1)
        self.assertEqual(sigpts.shape[1], self.ndim)

    def test_propagate_shape(self):
        sigpts = self.ut.sigma_points(randvars.Normal(self.mean, self.covar))
        propagated = self.ut.propagate(None, sigpts, lambda t, x: np.sin(x))
        self.assertEqual(propagated.ndim, 2)
        self.assertEqual(propagated.shape[0], 2 * self.ndim + 1)
        self.assertEqual(propagated.shape[1], self.ndim)

    def test_estimate_statistics_shape(self):
        sigpts = self.ut.sigma_points(randvars.Normal(self.mean, self.covar))
        proppts = self.ut.propagate(None, sigpts, lambda t, x: np.sin(x))
        mest, cest, ccest = self.ut.estimate_statistics(
            proppts, sigpts, self.covar, self.mean
        )
        self.assertEqual(mest.ndim, 1)
        self.assertEqual(mest.shape[0], self.ndim)
        self.assertEqual(cest.ndim, 2)
        self.assertEqual(cest.shape[0], self.ndim)
        self.assertEqual(cest.shape[1], self.ndim)
        self.assertEqual(ccest.ndim, 2)
        self.assertEqual(ccest.shape[0], self.ndim)
        self.assertEqual(ccest.shape[1], self.ndim)

    def test_transform_of_gaussian_exact(self):
        sigpts = self.ut.sigma_points(randvars.Normal(self.mean, self.covar))
        ndim_meas = self.ndim + 1  # != self.ndim is important
        transmtrx = np.random.rand(ndim_meas, self.ndim)
        meascov = 0 * np.random.rand(ndim_meas, ndim_meas)
        proppts = self.ut.propagate(None, sigpts, lambda t, x: transmtrx @ x)
        mest, cest, ccest = self.ut.estimate_statistics(
            proppts, sigpts, meascov, self.mean
        )
        diff_mean = np.linalg.norm(mest - transmtrx @ self.mean)
        diff_covar = np.linalg.norm(cest - transmtrx @ self.covar @ transmtrx.T)
        diff_crosscovar = np.linalg.norm(ccest - self.covar @ transmtrx.T)
        self.assertLess(diff_mean / np.linalg.norm(transmtrx @ self.mean), 1e-11)
        self.assertLess(
            diff_covar / np.linalg.norm(transmtrx @ self.covar @ transmtrx.T), 1e-11
        )
        self.assertLess(
            diff_crosscovar / np.linalg.norm(self.covar @ transmtrx.T), 1e-11
        )



"""Categorical random variables."""
from typing import Optional

import numpy as np

from probnum.type import RandomStateArgType

from ._random_variable import DiscreteRandomVariable


class Categorical(DiscreteRandomVariable):
    """Categorical random variable.

    Parameters
    ----------
    probabilities :
        Probabilities of the events.
    support :
        Support of the categorical distribution. Optional. Default is None,
        in which case the support is chosen as :math:`(0, ..., K-1)` where
        :math:`K` is the number of elements in `event_probabilities`.
    random_state :
        Random state of the random variable.
    """

    def __init__(
        self,
        probabilities: np.ndarray,
        support: Optional[np.ndarray] = None,
        random_state: Optional[RandomStateArgType] = None,
    ):
        # The set of events is names "support" to be aligned with the method
        # DiscreteRandomVariable.in_support().

        num_categories = len(probabilities)
        self._probabilities = np.asarray(probabilities)
        self._support = (
            np.asarray(support) if support is not None else np.arange(num_categories)
        )

        parameters = {
            "support": self._support,
            "probabilities": self._probabilities,
            "num_categories": num_categories,
        }

        def _sample_categorical(size=()):
            """Sample from a categorical distribution.

            While on first sight, one might think that this
            implementation can be replaced by
            `np.random.choice(self.support, size, self.probabilities)`,
            this is not true, because `np.random.choice` cannot handle
            arrays with `ndim > 1`, but `self.support` can be just that.
            This detour via the `mask` avoids this problem.
            """
            mask = np.random.choice(
                np.arange(len(self.support)), size=size, p=self.probabilities
            ).reshape(size)
            return self.support[mask]

        def _pmf_categorical(x):
            """PMF of a categorical distribution.

            This implementation is defense against cryptic warnings such as:
            # https://stackoverflow.com/questions/45020217/numpy-where-function-throws-a-futurewarning-returns-scalar-instead-of-list
            """
            x = np.asarray(x)
            if x.dtype != self.dtype:
                raise ValueError(
                    "The data type of x does not match with the data type of the support."
                )

            mask = (x == self.support).nonzero()[0]
            return self.probabilities[mask][0] if len(mask) > 0 else 0.0

        def _mode_categorical():
            mask = np.argmax(self.probabilities)
            return self.support[mask]

        super().__init__(
            shape=self._support[0].shape,
            dtype=self._support[0].dtype,
            random_state=random_state,
            parameters=parameters,
            sample=_sample_categorical,
            pmf=_pmf_categorical,
            mode=_mode_categorical,
        )

    @property
    def probabilities(self) -> np.ndarray:
        """Event probabilities of the categorical distribution."""
        return self._probabilities

    @property
    def support(self) -> np.ndarray:
        """Support of the categorical distribution."""
        return self._support

    def resample(self) -> "Categorical":
        """Resample the support of the categorical random variable.

        Return a new categorical random variable (RV), where the support
        is randomly chosen from the elements in the current support with
        probabilities given by the current event probabilities. The
        probabilities of the resulting categorical RV are all equal.
        """
        num_events = len(self.support)
        new_support = self.sample(size=num_events)
        new_probabilities = np.ones(self.probabilities.shape) / num_events
        return Categorical(
            support=new_support,
            probabilities=new_probabilities,
            random_state=self.random_state,
        )



import numpy as np
import pytest

import probnum.statespace as pnss
from probnum import randvars

from .test_transition import InterfaceTestTransition


class TestDiscreteGaussian(InterfaceTestTransition):
    """Tests for the discrete Gaussian class.

    Most of the tests are reused/overwritten in subclasses, therefore
    the class- structure. Unfortunately, testing the product space of
    all combinations (different subclasses, different implementations,
    etc.) generates a lot of tests that are executed.
    """

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(self, test_ndim, spdmat1):

        self.g = lambda t, x: np.sin(x)
        self.S = lambda t: spdmat1
        self.dg = lambda t, x: np.cos(x)
        self.transition = pnss.DiscreteGaussian(
            test_ndim,
            test_ndim,
            self.g,
            self.S,
            self.dg,
            None,
        )

    # Test access to system matrices

    def test_state_transition(self, some_normal_rv1):
        received = self.transition.state_trans_fun(0.0, some_normal_rv1.mean)
        expected = self.g(0.0, some_normal_rv1.mean)
        np.testing.assert_allclose(received, expected)

    def test_process_noise(self):
        received = self.transition.proc_noise_cov_mat_fun(0.0)
        expected = self.S(0.0)
        np.testing.assert_allclose(received, expected)

    def test_process_noise_cholesky(self):
        received = self.transition.proc_noise_cov_cholesky_fun(0.0)
        expected = np.linalg.cholesky(self.transition.proc_noise_cov_mat_fun(0.0))
        np.testing.assert_allclose(received, expected)

    def test_jacobian(self, some_normal_rv1):
        received = self.transition.jacob_state_trans_fun(0.0, some_normal_rv1.mean)
        expected = self.dg(0.0, some_normal_rv1.mean)
        np.testing.assert_allclose(received, expected)

    # Test forward and backward implementations

    def test_forward_rv(self, some_normal_rv1):
        with pytest.raises(NotImplementedError):
            self.transition.forward_rv(some_normal_rv1, 0.0)

    def test_forward_realization(self, some_normal_rv1):
        out, _ = self.transition.forward_realization(some_normal_rv1.sample(), 0.0)
        assert isinstance(out, randvars.Normal)

    def test_backward_rv(self, some_normal_rv1, some_normal_rv2):
        with pytest.raises(NotImplementedError):
            self.transition.backward_rv(some_normal_rv1, some_normal_rv2)

    def test_backward_realization(self, some_normal_rv1, some_normal_rv2):
        with pytest.raises(NotImplementedError):
            self.transition.backward_realization(
                some_normal_rv1.sample(), some_normal_rv2
            )

    def test_input_dim(self, test_ndim):
        assert self.transition.input_dim == test_ndim

    def test_output_dim(self, test_ndim):
        assert self.transition.output_dim == test_ndim


@pytest.fixture(params=["classic", "sqrt"])
def forw_impl_string_linear_gauss(request):
    """Forward implementation choices passed via strings."""
    return request.param


@pytest.fixture(params=["classic", "joseph", "sqrt"])
def backw_impl_string_linear_gauss(request):
    """Backward implementation choices passed via strings."""
    return request.param


class TestLinearGaussian(TestDiscreteGaussian):
    """Test class for linear Gaussians. Inherits tests from `TestDiscreteGaussian` but
    overwrites the forward and backward transitions.

    Also tests that different forward and backward implementations yield
    the same results.
    """

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(
        self,
        test_ndim,
        spdmat1,
        spdmat2,
        forw_impl_string_linear_gauss,
        backw_impl_string_linear_gauss,
    ):

        self.G = lambda t: spdmat1
        self.S = lambda t: spdmat2
        self.v = lambda t: np.arange(test_ndim)
        self.transition = pnss.DiscreteLinearGaussian(
            test_ndim,
            test_ndim,
            self.G,
            self.v,
            self.S,
            forward_implementation=forw_impl_string_linear_gauss,
            backward_implementation=backw_impl_string_linear_gauss,
        )

        self.g = lambda t, x: self.G(t) @ x + self.v(t)
        self.dg = lambda t, x: self.G(t)

    # Test access to system matrices

    def test_state_transition_mat_fun(self):
        received = self.transition.state_trans_mat_fun(0.0)
        expected = self.G(0.0)
        np.testing.assert_allclose(received, expected)

    def test_shift_vec_fun(self):
        received = self.transition.shift_vec_fun(0.0)
        expected = self.v(0.0)
        np.testing.assert_allclose(received, expected)

    # Test forward and backward implementations

    def test_forward_rv(self, some_normal_rv1):
        out, _ = self.transition.forward_rv(some_normal_rv1, 0.0)
        assert isinstance(out, randvars.Normal)

    def test_backward_rv(self, some_normal_rv1, some_normal_rv2):
        out, _ = self.transition.backward_rv(some_normal_rv1, some_normal_rv2)
        assert isinstance(out, randvars.Normal)

    def test_backward_realization(self, some_normal_rv1, some_normal_rv2):
        out, _ = self.transition.backward_realization(
            some_normal_rv1.sample(), some_normal_rv2
        )
        assert isinstance(out, randvars.Normal)

    def test_all_forward_rv_same(self, some_normal_rv1, diffusion):
        """Assert all implementations give the same output."""
        out_classic, info_classic = self.transition._forward_rv_classic(
            some_normal_rv1, 0.0, compute_gain=True, _diffusion=diffusion
        )
        out_sqrt, info_sqrt = self.transition._forward_rv_sqrt(
            some_normal_rv1, 0.0, compute_gain=True, _diffusion=diffusion
        )

        np.testing.assert_allclose(out_classic.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_classic.cov, out_sqrt.cov)
        np.testing.assert_allclose(info_classic["crosscov"], info_sqrt["crosscov"])
        np.testing.assert_allclose(info_classic["gain"], info_sqrt["gain"])

    def test_all_backward_rv_same_no_cache(
        self, some_normal_rv1, some_normal_rv2, diffusion
    ):
        """Assert all implementations give the same output -- no gain or forwarded RV
        passed."""

        out_classic, _ = self.transition._backward_rv_classic(
            some_normal_rv1, some_normal_rv2, t=0.0, _diffusion=diffusion
        )
        out_sqrt, _ = self.transition._backward_rv_sqrt(
            some_normal_rv1, some_normal_rv2, t=0.0, _diffusion=diffusion
        )
        out_joseph, _ = self.transition._backward_rv_joseph(
            some_normal_rv1, some_normal_rv2, t=0.0, _diffusion=diffusion
        )

        # Classic -- sqrt
        np.testing.assert_allclose(out_classic.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_classic.cov, out_sqrt.cov)

        # Joseph -- sqrt
        np.testing.assert_allclose(out_joseph.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_joseph.cov, out_sqrt.cov)

    def test_all_backward_rv_same_with_cache(
        self, some_normal_rv1, some_normal_rv2, diffusion
    ):
        """Assert all implementations give the same output -- gain and forwarded RV
        passed."""

        rv_forward, info = self.transition.forward_rv(
            some_normal_rv2, 0.0, compute_gain=True, _diffusion=diffusion
        )
        gain = info["gain"]

        out_classic, _ = self.transition._backward_rv_classic(
            some_normal_rv1,
            some_normal_rv2,
            rv_forwarded=rv_forward,
            gain=gain,
            t=0.0,
            _diffusion=diffusion,
        )
        out_sqrt, _ = self.transition._backward_rv_sqrt(
            some_normal_rv1,
            some_normal_rv2,
            rv_forwarded=rv_forward,
            gain=gain,
            t=0.0,
            _diffusion=diffusion,
        )
        out_joseph, _ = self.transition._backward_rv_joseph(
            some_normal_rv1,
            some_normal_rv2,
            rv_forwarded=rv_forward,
            gain=gain,
            t=0.0,
            _diffusion=diffusion,
        )

        # Classic -- sqrt
        np.testing.assert_allclose(out_classic.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_classic.cov, out_sqrt.cov)

        # Joseph -- sqrt
        np.testing.assert_allclose(out_joseph.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_joseph.cov, out_sqrt.cov)

    def test_all_backward_realization_same_no_cache(
        self, some_normal_rv1, some_normal_rv2, diffusion
    ):
        """Assert all implementations give the same output -- no gain or forwarded RV
        passed."""

        out_classic, _ = self.transition._backward_rv_classic(
            randvars.Constant(some_normal_rv1.mean),
            some_normal_rv2,
            t=0.0,
            _diffusion=diffusion,
        )
        out_sqrt, _ = self.transition._backward_rv_sqrt(
            randvars.Constant(some_normal_rv1.mean),
            some_normal_rv2,
            t=0.0,
            _diffusion=diffusion,
        )
        out_joseph, _ = self.transition._backward_rv_joseph(
            randvars.Constant(some_normal_rv1.mean),
            some_normal_rv2,
            t=0.0,
            _diffusion=diffusion,
        )

        # Classic -- sqrt
        np.testing.assert_allclose(out_classic.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_classic.cov, out_sqrt.cov)

        # Joseph -- sqrt
        np.testing.assert_allclose(out_joseph.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_joseph.cov, out_sqrt.cov)

    def test_all_backward_realization_same_with_cache(
        self, some_normal_rv1, some_normal_rv2, diffusion
    ):
        """Assert all implementations give the same output -- gain and forwarded RV
        passed."""

        rv_forward, info = self.transition.forward_rv(
            some_normal_rv2, 0.0, compute_gain=True, _diffusion=diffusion
        )
        gain = info["gain"]

        out_classic, _ = self.transition._backward_rv_classic(
            randvars.Constant(some_normal_rv1.mean),
            some_normal_rv2,
            rv_forwarded=rv_forward,
            gain=gain,
            t=0.0,
            _diffusion=diffusion,
        )
        out_sqrt, _ = self.transition._backward_rv_sqrt(
            randvars.Constant(some_normal_rv1.mean),
            some_normal_rv2,
            rv_forwarded=rv_forward,
            gain=gain,
            t=0.0,
            _diffusion=diffusion,
        )
        out_joseph, _ = self.transition._backward_rv_joseph(
            randvars.Constant(some_normal_rv1.mean),
            some_normal_rv2,
            rv_forwarded=rv_forward,
            gain=gain,
            t=0.0,
            _diffusion=diffusion,
        )

        # Classic -- sqrt
        np.testing.assert_allclose(out_classic.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_classic.cov, out_sqrt.cov)

        # Joseph -- sqrt
        np.testing.assert_allclose(out_joseph.mean, out_sqrt.mean)
        np.testing.assert_allclose(out_joseph.cov, out_sqrt.cov)


class TestLTIGaussian(TestLinearGaussian):

    # Replacement for an __init__ in the pytest language. See:
    # https://stackoverflow.com/questions/21430900/py-test-skips-test-class-if-constructor-is-defined
    @pytest.fixture(autouse=True)
    def _setup(
        self,
        test_ndim,
        spdmat1,
        spdmat2,
        forw_impl_string_linear_gauss,
        backw_impl_string_linear_gauss,
    ):

        self.G_const = spdmat1
        self.S_const = spdmat2
        self.v_const = np.arange(test_ndim)
        self.transition = pnss.DiscreteLTIGaussian(
            self.G_const,
            self.v_const,
            self.S_const,
            forward_implementation=forw_impl_string_linear_gauss,
            backward_implementation=backw_impl_string_linear_gauss,
        )

        # Compatibility with superclass' test
        self.G = lambda t: self.G_const
        self.S = lambda t: self.S_const
        self.v = lambda t: self.v_const
        self.g = lambda t, x: self.G(t) @ x + self.v(t)
        self.dg = lambda t, x: self.G(t)

    # Test access to system matrices

    def test_state_transition_mat(self):
        received = self.transition.state_trans_mat
        expected = self.G_const
        np.testing.assert_allclose(received, expected)

    def test_shift_vec(self):
        received = self.transition.shift_vec
        expected = self.v_const
        np.testing.assert_allclose(received, expected)

    def test_process_noise_cov_mat(self):
        received = self.transition.proc_noise_cov_mat
        expected = self.S_const
        np.testing.assert_allclose(received, expected)

    def test_process_noise_cov_cholesky(self):
        received = self.transition.proc_noise_cov_cholesky
        expected = np.linalg.cholesky(self.S_const)
        np.testing.assert_allclose(received, expected)



"""Randomly draw nodes from the measure to use for integration."""

import numpy as np

from probnum.quad._integration_measures import IntegrationMeasure


def sample_from_measure(nevals: int, measure: IntegrationMeasure) -> np.ndarray:
    r"""Acquisition policy: Draw random samples from the integration measure.

    Parameters
    ----------
    nevals : int
        Number of function evaluations.

    measure :
        The integration measure :math:`\mu`.

    Returns
    -------
    x : np.ndarray
        Nodes where the integrand will be evaluated.
    """
    return measure.sample(nevals).reshape(nevals, -1)



"""Definitions and collection of problems solved by probabilistic numerical methods."""

from ._problems import (
    InitialValueProblem,
    LinearSystem,
    QuadratureProblem,
    RegressionProblem,
)

__all__ = [
    "RegressionProblem",
    "InitialValueProblem",
    "LinearSystem",
    "QuadratureProblem",
]

# Set correct module paths. Corrects links and module paths in documentation.
RegressionProblem.__module__ = "probnum.problems"
InitialValueProblem.__module__ = "probnum.problems"
LinearSystem.__module__ = "probnum.problems"
QuadratureProblem.__module__ = "probnum.problems"
